#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#ifdef we32000

#include <limits.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/times.h>

#define MAX_SIZE	262144

#else

#include <sys/time.h>
#include <sys/resource.h>

#ifdef __ultrix__
# define MAX_SIZE	1048576
#else
# define MAX_SIZE	4194304
#endif

#endif

#pragma GCC optimize("O3")

void downHeap(int *a, int k, int n)
{
	int i, tmp;

	n--;
	for (tmp = a[k]; k <= n / 2; k = i) {
		i = 2 * k;
		if (i < n && a[i] < a[i+1])
			i++;
		if (tmp >= a[i])
			break;
		a[k] = a[i];
	}
	a[k] = tmp;
}

void heapSortC(int *a, int n)
{
	int i, tmp;

	for (i = n / 2; i >= 0; i--)
		downHeap(a, i, n);
	for (i = n - 1; i > 0; i--) {
		tmp = a[i];
		a[i] = a[0];
		a[0] = tmp;
		downHeap(a, 0, i);
	}
}

#ifdef __i386__
# define ATTRIB	__attribute__((regparm(3)))
#else
# define ATTRIB
#endif

void ATTRIB heapSort(int *, int);

int comp(const void *a, const void *b)
{
	return *(int *)a - *(int *)b;
}

long long useconds()
{
#ifdef we32000
	struct tms t;

	if (times(&t) == -1) {
		perror("times()");
		exit(errno);
	}
	return (long long)(t.tms_utime * (1000000 / CLK_TCK));
#else
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru)) {
		perror("getrusage()");
		exit(errno);
	}
	return (long long)(ru.ru_utime.tv_sec + ru.ru_stime.tv_sec) * 1000000LL
			+ ru.ru_utime.tv_usec + ru.ru_stime.tv_usec;
#endif
}

long long fill_in_and_start(int *a, int n)
{
	int i;

	srand48(1);
	for (i = 0; i < n; i++)
		a[i] = mrand48();
	return useconds();
}

void print_time(const char *s, long long old)
{
	printf("%s: %.6f s\n", s, (double)(useconds() - old) / 1000000.);
}

int main()
{
	int i, size, *a = malloc(MAX_SIZE * sizeof(int));
	long long start;

	for (i = 0, size = MAX_SIZE / 1024; i < 11; i++, size *= 2) {
		printf("Array size %7d:", size);
		start = fill_in_and_start(a, size);
		qsort(a, size, sizeof(int), comp);
		print_time("\tQuickSort", start);
		start = fill_in_and_start(a, size);
		heapSortC(a, size);
		print_time("\t\t\tHeapSortC", start);
		start = fill_in_and_start(a, size);
		heapSort(a, size);
		print_time("\t\t\t HeapSort", start);
	}
	free(a);
	return 0;
}
