#include <inttypes.h>

uint64_t udiv128(uint64_t a_lo, uint64_t a_hi, uint64_t b, uint64_t *rem);

uint64_t Udiv128(uint64_t a_lo, uint64_t a_hi, uint64_t b, uint64_t *rem)
{
	uint64_t q, r, carry, temp;
	int i;

	q = a_lo << 1;	// http://codereview.stackexchange.com/questions/67962
	r = a_hi;
	carry = a_lo >> 63;
	for (i = 0; i < 64; i++) {
		temp = r >> 63;
		r = r << 1 | carry;
		carry = temp;
		if (carry == 0) {
			if (r < b)
				goto shift;
			carry++;
		}
		r -= b;
	shift:	temp = q >> 63;
		q = q << 1 | carry;
		carry = temp;
	}
	*rem = r;
	return q;
}

#include <stdio.h>

int main(void)	// http://ttmath.org/online_calculator
{
	uint64_t tmp, high = 0, low = 0, divisor, quotient, remainder;
	char c, s[41];	// 2^128 = 340282366920938463463374607431768211456
	int r1, r2, i;	//	 0^	  10^	    20^	      30^	40^

	setvbuf(stdout, NULL, _IONBF, 0);
	printf("  Dividend: ");
	r1 = scanf("%40s", s);
	for (i = 0; (c = s[i]) >= '0' && c <= '9'; i++) {
		high *= 10;	// http://stackoverflow.com/questions/37756465
		high += ((low >> 32) * 10 + ((low & UINT32_MAX) * 10 >> 32)) >> 32;
		low *= 10;
		tmp = low;
		low += c - '0';
		high += low < tmp;
	}
	printf("   Divisor: ");
	r2 = scanf("%llu", &divisor);
	// > 340282366920938463444927863358058659839 / 18446744073709551615
	if (high >= divisor) {
		printf("Overflow!\n");
		return -1;
	}
	quotient = Udiv128(low, high, divisor, &remainder);
	printf("\nC quotient: %llu, remainder: %llu", quotient, remainder);
#if __WORDSIZE == 64
	__uint128_t dividend = (__uint128_t)high << 64 | low;
	remainder = (uint64_t)(dividend % divisor);
	quotient = (uint64_t)(dividend / divisor);
#else
	remainder = UINT64_MAX;
	quotient = udiv128(low, high, divisor, &remainder);
#endif
	printf("\n  Quotient: %llu, remainder: %llu\n", quotient, remainder);
	return r1 - r2;
}
